"""This script has two purposes:

1. Apply the attributes from a cobra model to an existing Escher
map. For instance, update modified reversibilities.

2. Convert maps made with Escher beta versions to valid jsonschema/1-0-0
maps.

Usage:

Using a Theseus model name (https://github.com/zakandrewking/Theseus).
python convert_map.py old_map.json iJO1366

Using a SBML or JSON model path
python convert_map.py another_old_map.json path/to/iJO1366.sbml

"""

from __future__ import print_function, unicode_literals

try:
    import cobra.io
    import jsonschema
except ImportError:
    raise Exception(('The Python packages jsonschema and COBRApy (0.3.0b3 or later) '
                     'are required for converting maps.'))
try:
    from theseus import load_model
except ImportError:
    print('Theseus not available (https://github.com/zakandrewking/Theseus)')
import sys
import json
from os.path import basename, join
try:
    from urllib.request import urlopen
except ImportError:
    from urllib import urlopen

from escher.validate import validate_map, genes_for_gene_reaction_rule

def main():
    """Load an old Escher map, and generate a validated map.

    """
    try:
        in_file = sys.argv[1]
        model_path = sys.argv[2]
    except IndexError:
        raise Exception("Usage: python convert_map.py old_map.json path/to/model.sbml")
   
    # get the cobra model
    try:
        model = load_model(model_path)
    except Exception:
        try:
            model = cobra.io.load_json_model(model_path)
        except (IOError, ValueError):
            try:
                model = cobra.io.read_sbml_model(model_path)
            except IOError:
                raise Exception('Could not find model in theseus or filesystem: %s' % model_path)

    # get the current map
    with open(in_file, 'r') as f:
        out = json.load(f)

    the_map = convert(out, model)
    
    # don't replace the file
    out_file = in_file.replace('.json', '_converted.json')
    print('Saving validated map to %s' % out_file)
    with open(out_file, 'w') as f:
        json.dump(the_map, f, allow_nan=False)

def convert(map, model):
    # check for new, 2-level maps
    has_head = isinstance(map, list) and len(map) == 2
    if has_head:
        map_body = map[1]
    else:
        map_body = map

    # add missing elements
    for k in ['nodes', 'reactions', 'text_labels']:
        if k not in map_body or len(map_body[k]) == 0:
            map_body[k] = {}
    # default canvas
    if 'canvas' not in map_body:
        map_body['canvas'] = { 'x': -1440,
                               'y': -775,
                               'width': 4320,
                               'height': 2325 }        
    
    # keep track of deleted nodes and reactions
    nodes_to_delete = set()
    reactions_to_delete = set()
    nodes_with_segments = set()

    # nodes
    for id, node in map_body['nodes'].items():
        # follow rules for type
        # metabolites
        if node['node_type'] == 'metabolite':
            # no bigg_id
            if not 'bigg_id' in node:
                print('No bigg_id for node %s. Deleting.' % id)
                nodes_to_delete.add(id)
                
            # unsupported attributes
            for key in list(node.keys()):
                if not key in ["node_type", "x", "y", "bigg_id", "name",
                               "label_x", "label_y", "node_is_primary"]:
                    del node[key]
                    
            # find the metabolite
            try:
                cobra_metabolite = model.metabolites.get_by_id(node['bigg_id'])
            except KeyError:
                print('Could not find metabolite %s in model. Deleting.' % node['bigg_id'])
                nodes_to_delete.add(id)
                continue
            # apply new display names
            node['name'] = cobra_metabolite.name

            # node_is_primary defaults to False
            if not 'node_is_primary' in node or node['node_is_primary'] not in [True, False]:
                node['node_is_primary'] = False
        # markers
        elif node['node_type'] in ['multimarker', 'midmarker']:
            # unsupported attributes
            for key in list(node.keys()):
                if not key in ["node_type", "x", "y"]:
                    del node[key]
        # invalid node_type
        else:
            nodes_to_delete.add(id)
        
    # update reactions
    for id, reaction in map_body['reactions'].items():
        # missing attributes
        will_delete = False
        for k in ["bigg_id", "segments", "label_x", "label_y"]:
            if not k in reaction:
                print('No %s for reaction %s. Deleting.' % (k, id))
                reactions_to_delete.add(id)
                will_delete = True
        if will_delete:
            continue

        # unsupported attributes
        for key in list(reaction.keys()):
            if not key in ["name", "bigg_id","reversibility",
                           "label_x", "label_y", "gene_reaction_rule",
                           "genes", "metabolites", "segments"]:
                del reaction[key]

        # cast attributes
        for k in ['label_x', 'label_y']:
            reaction[k] = float(reaction[k])

        # unsupported attributes in segments
        for s_id, segment in reaction['segments'].items():
            for key in list(segment.keys()):
                if not key in ["from_node_id", "to_node_id", "b1", "b2"]:
                    del segment[key]

            # cast attributes
            for k in ['b1', 'b2']:
                if segment[k] is not None and (segment[k]['x'] is None or segment[k]['y'] is None):
                    segment[k] = None

            # keep track of nodes that have appeared here
            for key in ['from_node_id', 'to_node_id']:
                try:
                    nodes_with_segments.add(segment[key])
                except KeyError:
                    pass

        # get reaction
        try:
            cobra_reaction = model.reactions.get_by_id(reaction['bigg_id'])
        except KeyError:
            print('Could not find reaction %s in model. Deleting.' % reaction['bigg_id'])
            reactions_to_delete.add(id)
            continue
        reaction['gene_reaction_rule'] = cobra_reaction.gene_reaction_rule
        reaction['reversibility'] = (cobra_reaction.lower_bound < 0 and cobra_reaction.upper_bound > 0)
        # reverse metabolites if reaction runs in reverse
        rev_mult = (-1 if
                    (cobra_reaction.lower_bound < 0 and cobra_reaction.upper_bound <= 0)
                    else 1)
        # use metabolites from reaction
        reaction['metabolites'] = [{'bigg_id': met.id, 'coefficient': coeff * rev_mult}
                                   for met, coeff in
                                   cobra_reaction.metabolites.items()]
            
        reaction['name'] = cobra_reaction.name

        reaction['genes'] = []
        for gene in genes_for_gene_reaction_rule(reaction['gene_reaction_rule']):
            try:
                cobra_gene = model.genes.get_by_id(gene)
            except KeyError:
                print('Could not find gene %s in model. Setting name to ID.')
                reaction['genes'].append({'bigg_id': gene, 'name': gene})
                continue
            reaction['genes'].append({'bigg_id': gene, 'name': cobra_gene.name})
        
        # remove any lost segments
        reaction['segments'] = {id: seg for id, seg in reaction['segments'].items()
                                if ((seg['to_node_id'] in map_body['nodes'] and seg['to_node_id'] not in nodes_to_delete) and
                                    (seg['from_node_id'] in map_body['nodes'] and seg['from_node_id'] not in nodes_to_delete))}

    # delete those reactions
    for reaction_id in reactions_to_delete:
        del map_body['reactions'][reaction_id]

    # delete those nodes
    for node_id in nodes_to_delete:
        del map_body['nodes'][node_id]

    # delete any nodes with no segment
    for node_id in list(map_body['nodes'].keys()):
        if node_id not in nodes_with_segments:
            # may not be there, if previously deleted
            try:
                del map_body['nodes'][node_id]
                print('No segments for node %s. Deleting' % node_id)
            except KeyError:
                pass
        
    # text labels
    for id, text_label in map_body['text_labels'].items():
        # unsupported attributes
        for key in list(text_label.keys()):
            if not key in ["x", "y", "text"]:
                del text_label[key]

    # canvas
    # unsupported attributes
    for key in list(map_body['canvas'].keys()):
        if not key in ["x", "y", "width", "height"]:
            del map_body['canvas'][key]

    # delete unsupported elements
    for key in list(map_body.keys()):
        if not key in ["nodes", "reactions", "text_labels", "canvas"]:
            del map_body[key]

    header = {
        "schema": "https://escher.github.io/escher/jsonschema/1-0-0#",
        "homepage": "https://escher.github.io",
        "map_name": "",
        "map_id": "",
        "map_description": ""
        }
    if has_head:
        for key, value in map[0].items():
            if key in ['schema', 'homepage']: continue
            header[key] = value
    
    the_map = [header, map_body]

    validate_map(the_map) 
    return the_map

if __name__=="__main__":
    main()



# -*- coding: utf-8 -*-

from __future__ import print_function, unicode_literals

from escher.version import __version__, __schema_version__
import os
import re
from os.path import dirname, realpath, join

root_directory = realpath(join(dirname(__file__), '..'))

_escher_local = {
    'builder_embed_css': 'escher/css/builder-embed-%s.css' % __version__,
    'builder_css': 'escher/css/builder-%s.css' % __version__,
    'escher': 'escher/lib/escher-%s.js' % __version__,
    'escher_min': 'escher/lib/escher-%s.min.js' % __version__,
    'logo': 'escher/resources/escher-logo@2x.png',
    'favicon': 'escher/resources/favicon.ico',
    'index_js': 'escher/js/web/index.js',
    'index_gh_pages_js': 'escher/js/web/index_gh_pages.js',
    'index_css': 'escher/css/web/index.css',
    'server_index': '%s/index.json' % __schema_version__,
    'map_download': '%s/maps/' % __schema_version__,
    'model_download': '%s/models/' % __schema_version__,
    }

_escher_web = {
    'builder_embed_css': 'builder-embed-%s.css' % __version__,
    'builder_css': 'builder-%s.css' % __version__,
    'escher': 'escher-%s.js' % __version__,
    'escher_min': 'escher-%s.min.js' % __version__,
    'server_index': '%s/index.json' % __schema_version__,
    'map_download': '%s/maps/' % __schema_version__,
    'model_download': '%s/models/' % __schema_version__,
    'favicon': 'escher/resources/favicon.ico',
    }
    
_dependencies = {
    'd3': 'escher/lib/d3.min.js',
    'boot_js': 'escher/lib/bootstrap-3.1.1.min.js',
    'boot_css': 'escher/lib/bootstrap-3.1.1.min.css',
    'jquery': 'escher/lib/jquery-2.1.0.min.js',
    'require_js': 'escher/lib/require.min.js',
    'bacon': 'escher/lib/bacon-0.7.12.min.js',
    }
    
_dependencies_cdn = {
    'd3': '//cdnjs.cloudflare.com/ajax/libs/d3/3.4.13/d3.min.js',
    'boot_js': '//netdna.bootstrapcdn.com/bootstrap/3.1.1/js/bootstrap.min.js',
    'boot_css': '//netdna.bootstrapcdn.com/bootswatch/3.1.1/simplex/bootstrap.min.css',
    'jquery': '//cdnjs.cloudflare.com/ajax/libs/jquery/2.1.0/jquery.min.js',
    'require_js': '//cdnjs.cloudflare.com/ajax/libs/require.js/2.1.11/require.min.js',
    'bacon': '//cdnjs.cloudflare.com/ajax/libs/bacon.js/0.7.12/bacon.min.js',
    }

_links = {
    'escher_root': '//escher.github.io/',
    'github': '//github.com/zakandrewking/escher/',
    'documentation': '//escher.readthedocs.org/',
    }

# external dependencies
names = list(_escher_local.keys()) + list(_escher_web.keys()) + list(_dependencies.keys()) + list(_links.keys())

def get_url(name, source='web', local_host=None, protocol=None):
    """Get a url.

    Arguments
    ---------

    name: The name of the URL. Options are available in urls.names.

    source: Either 'web' or 'local'. Cannot be 'local' for external links.

    protocol: The protocol can be 'http', 'https', or None which indicates a
    'protocol relative URL', as in //escher.github.io. Ignored if source is
    local.

    local_host: A host url, including the protocol. e.g. http://localhost:7778.

    """
    if source not in ['web', 'local']:
        raise Exception('Bad source: %s' % source)
    
    if protocol not in [None, 'http', 'https']:
        raise Exception('Bad protocol: %s' % protocol)

    if protocol is None:
        protocol = ''
    else:
        protocol = protocol + ':'

    def apply_local_host(url):
        return '/'.join([local_host.rstrip('/'), url.lstrip('/')])
        
    # escher
    if name in _escher_local and source == 'local':
        if local_host is not None:
            return apply_local_host(_escher_local[name])
        return _escher_local[name]
    elif name in _escher_web and source == 'web':
        return protocol + '/'.join([_links['escher_root'].rstrip('/'),
                                    _escher_web[name].lstrip('/')])
    # links
    elif name in _links:
        if source=='local':
            raise Exception('Source cannot be "local" for external links')
        return protocol + _links[name]
    # local dependencies
    elif name in _dependencies and source=='local':
        if local_host is not None:
            return apply_local_host(_dependencies[name])
        return _dependencies[name]
    # cdn dependencies
    elif name in _dependencies_cdn and source=='web':
        return protocol + _dependencies_cdn[name]
    
    raise Exception('name not found')



from theseus import load_model
import math
import cobra.io
import random
import json
import pandas as pd

# model
ijo = load_model('iJO1366')
ijo.genes.get_by_id('b1779').name = 'gapA'
cobra.io.save_json_model(ijo, 'iJO1366.json')

# gene data
gene_data = {gene.id: random.random() * 20 for gene in ijo.genes}
with open('gene_data_iJO1366.json', 'w') as f:
    json.dump(gene_data, f)

(pd.DataFrame.from_records(gene_data.items(), columns=['gene', 'value'])
 .to_csv('gene_data_iJO1366.csv', index=None))

# reaction text data
reaction_text_data = {reaction.id: reaction.build_reaction_string() for reaction in ijo.reactions}

(pd.DataFrame.from_records(reaction_text_data.items(), columns=['reaction', 'value'])
 .to_csv('reaction_text_data_iJO1366.csv', index=None))

# convert RNA-seq data to normalize data that looks like array data
with open('aerobic_anaerobic_E_coli_RNA-seq.json', 'r') as f:
    gene_comparison = json.load(f)
all_vals = gene_comparison[0].values() + gene_comparison[1].values()
# log values
all_vals = [math.pow(x, 1./4) for x in all_vals if x > 0]
the_min, the_max = min(all_vals), max(all_vals)
diff = the_max - the_min
norm = lambda x: (x * 2 / diff) - 1 - (the_min * 2 / diff)
out = []
for dataset in gene_comparison:
    this_out = {}
    for k, v in dataset.iteritems():
        if v > 0:
            this_out[k] = norm(math.pow(v, 1./4))
    out.append(this_out)
with open('aerobic_anaerobic_E_coli_RNA-seq_negatives.json', 'w') as f:
    json.dump(out, f)



# -*- coding: utf-8 -*-

from __future__ import print_function, unicode_literals

from escher.plots import (Builder, local_index, model_json_for_name,
                          map_json_for_name)
from escher.urls import get_url
from escher.urls import root_directory

import os, subprocess
from os.path import join
import tornado.ioloop
from tornado.web import RequestHandler, HTTPError, Application, asynchronous
from tornado.httpclient import AsyncHTTPClient
from tornado import gen
import tornado.escape
from tornado.options import define, options, parse_command_line
import json
import re
from jinja2 import Environment, PackageLoader
from mimetypes import guess_type

from escher.version import __version__, __schema_version__

# set up jinja2 template location
env = Environment(loader=PackageLoader('escher', 'templates'))

# set directory to server
NO_CACHE = False
PORT = 7778
PUBLIC = False
CAN_DEV = os.path.exists(join(root_directory, '.can_dev'))

def run(port=PORT, public=PUBLIC):
    global PORT
    global PUBLIC
    PORT = port
    PUBLIC = public
    print('serving directory %s on port %d' % (root_directory, PORT))
    application.listen(port, None if public else "localhost")
    try:
        tornado.ioloop.IOLoop.instance().start()
    except KeyboardInterrupt:
        print("bye!")

def stop():
    tornado.ioloop.IOLoop.instance().stop()

class BaseHandler(RequestHandler):
    def serve_path(self, path):
        # make sure the path exists
        if not os.path.isfile(path):
            raise HTTPError(404)
        # serve any raw file type
        with open(path, "rb") as file:
            data = file.read()
        # set the mimetype
        the_type = guess_type(path, strict=False)[0]
        self.set_header("Content-Type", ("application/octet-stream"
                                         if the_type is None
                                         else the_type))
        self.serve(data)
        
    def serve(self, data):
        if (NO_CACHE):
            self.set_header('Cache-Control', 'no-store, no-cache, must-revalidate, max-age=0')
        self.set_header('Access-Control-Allow-Origin', '*')
        self.write(data)
        self.finish()

class IndexHandler(BaseHandler):
    @asynchronous
    @gen.coroutine
    def get(self):
        # get the organisms, maps, and models
        response = yield gen.Task(AsyncHTTPClient().fetch, get_url('server_index', protocol='http'))
        if response.code == 200 and response.body is not None:
            server_index = response.body.decode('utf-8')
        else:
            server_index = json.dumps(None)

        # get the cached maps and models
        index = local_index()

        # render the template
        template = env.get_template('index.html')
        data = template.render(d3=get_url('d3', 'local'),
                               boot_css=get_url('boot_css', 'local'),
                               index_css=get_url('index_css', 'local'),
                               favicon=get_url('favicon', 'local'),
                               logo=get_url('logo', 'local'),
                               documentation=get_url('documentation', protocol='https'),
                               github=get_url('github'),
index_js=get_url('index_js', 'local'),
                               index_gh_pages_js=get_url('index_gh_pages_js', 'local'),
                               map_download=get_url('map_download', 'local'),
                               # server_index_url=
                               server_index=server_index,
                               local_index=json.dumps(index),
                               can_dev=CAN_DEV,
                               can_dev_js=json.dumps(CAN_DEV),
                               version=__version__,
                               web_version=False)
        
        self.set_header("Content-Type", "text/html")
        self.serve(data)
  
class BuilderHandler(BaseHandler):
    @asynchronous
    @gen.coroutine
    def get(self, kind, path):
        # builder vs. viewer
        enable_editing = (kind=='builder')

        # Builder options
        builder_kwargs = {}
        for a in ['starting_reaction', 'model_name', 'map_name', 'map_json',
                  'reaction_no_data_color', 'reaction_no_data_size',
                  'metabolite_no_data_color', 'metabolite_no_data_size',
                  'hide_secondary_nodes']:
            args = self.get_arguments(a)
            if len(args)==1:
                builder_kwargs[a] = (True if args[0].lower()=='true' else
                                     (False if args[0].lower()=='false' else
                                      args[0]))
        # array args
        for a in ['quick_jump', 'metabolite_size_range', 'metabolite_color_range',
                  'reaction_size_range', 'reaction_color_range', 'gene_styles']:
            args = self.get_arguments(a + '[]')
            if len(args) > 0:
                builder_kwargs[a] = args

        # js source
        args = self.get_arguments('js_source')
        js_source = args[0] if len(args) == 1 else 'web'

        # if the server is running locally, then the embedded css must be loaded
        # asynchronously using the same server thread.
        if js_source in ['dev', 'local']:  
            global PORT
            url = get_url('builder_embed_css',
                          source='local',
                          local_host='http://localhost:%d' % PORT)
            response = yield gen.Task(AsyncHTTPClient().fetch, url)
            if response.code != 200 or response.body is None:
                raise Exception('Could not load embedded_css from %s' % url)

            builder_kwargs['embedded_css'] = (response.body
                                              .decode('utf-8')
                                              .replace('\n', ' '))

        # example data
        def load_data_file(rel_path):
            """Load a JSON file with relative path."""
            try:
                with open(join(root_directory, rel_path), 'r') as f:
                    return json.load(f)
            except:
                logging.warn('Could not load example_data file: %s' % rel_path)
        if len(self.get_arguments('example_data')) > 0:
            r_filepath = 'escher/example_data/reaction_data_iJO1366.json'
            builder_kwargs['reaction_data'] = load_data_file(r_filepath)
            m_filepath = 'escher/example_data/metabolite_data_iJO1366.json'
            builder_kwargs['metabolite_data'] = load_data_file(m_filepath)
        
        # make the builder
        builder = Builder(safe=True, **builder_kwargs)
            
        # display options
        display_kwargs = {'minified_js': True,
                          'scroll_behavior': 'pan',
                          'menu': 'all'}
            
        # keyword
        for a in ['menu', 'scroll_behavior', 'minified_js',
                  'auto_set_data_domain', 'never_ask_before_quit']:
            args = self.get_arguments(a)
            if len(args)==1:
                display_kwargs[a] = (True if args[0].lower()=='true' else
                                     (False if args[0].lower()=='false' else
                                      args[0]))

        html = builder._get_html(js_source=js_source, enable_editing=enable_editing,
                                 enable_keys=True, html_wrapper=True, fill_screen=True,
                                 height='100%', **display_kwargs)
        
        self.set_header("Content-Type", "text/html")
        self.serve(html)

class MapModelHandler(BaseHandler):
    def get(self, path):
        try:
            kind, organism, name = path.strip('/').split('/')
        except (TypeError, ValueError):
            raise Exception('invalid path %s' % path)
        if kind == 'maps':
            b = Builder(map_name=name)
            self.set_header('Content-Type', 'application/json')
            self.serve(b.loaded_map_json)
        else:
            b = Builder(model_name=name)
            self.set_header('Content-Type', 'application/json')
            self.serve(b.loaded_model_json)
         
class StaticHandler(BaseHandler):
    def get(self, path):
        path = join(root_directory, path)
        print('getting path %s' % path)
        self.serve_path(path)

class DocsHandler(BaseHandler):
    def get(self, path):
        path = join(root_directory, 'docs', '_build', 'html', path)
        print('getting path %s' % path)
        self.serve_path(path)

settings = {"debug": "False"}

application = Application([
    (r"/(escher/lib/.*)", StaticHandler),
    (r"/(escher/fonts/.*)", StaticHandler),
    (r"/(escher/js/.*)", StaticHandler),
    (r"/(escher/css/.*)", StaticHandler),
    (r"/(escher/resources/.*)", StaticHandler),
    (r"/(escher/jsonschema/.*)", StaticHandler),
    (r"/(builder|viewer)(.*)", BuilderHandler),
    (r"/%s(/.*)" % __schema_version__, MapModelHandler),
    (r"/docs/(.*)", DocsHandler),
    (r"/", IndexHandler),
], **settings)
 
if __name__ == "__main__":
    # define port
    define("port", default=PORT, type=int, help="Port to serve on.")
    define("public", default=PUBLIC, type=bool,
           help=("If False, listen only on localhost. If True, listen on "
                 "all available addresses."))
    parse_command_line()
    run(port=options.port, public=options.public)



""" The following code has been adapted from mpld3. Modifications (c) 2014,
Zachary King.

mpld3, http://mpld3.github.io/, A Simple server used to show mpld3 images.
Copyright (c) 2013, Jake Vanderplas
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice, this
  list of conditions and the following disclaimer in the documentation and/or
  other materials provided with the distribution.

* Neither the name of the {organization} nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
import sys
import threading
import webbrowser
import socket
import itertools
import random

IPYTHON_WARNING = """
Note: You must interrupt the kernel to end this command
"""

try:
    # Python 2.x
    import BaseHTTPServer as server
except ImportError:
    # Python 3.x
    from http import server

def generate_handler(html, files=None):
    if files is None:
        files = {}

    class MyHandler(server.BaseHTTPRequestHandler):
        def do_GET(self):
            """Respond to a GET request."""
            if self.path == '/':
                self.send_response(200)
                self.send_header("Content-type", "text/html")
                self.end_headers()
                self.wfile.write('<!DOCTYPE html>'
                                 '<html lang="en"><head>'
                                 '<title>Escher map</title>'
                                 '</head><body>\n')
                self.wfile.write(html)
                self.wfile.write("</body></html>")
            elif self.path in files:
                content_type, content = files[self.path]
                self.send_response(200)
                self.send_header("Content-type", content_type)
                self.end_headers()
                self.wfile.write(content)
            else:
                self.send_error(404)

    return MyHandler

def find_open_port(ip, port, n=50):
    """Find an open port near the specified port"""
    ports = itertools.chain((port + i for i in range(n)),
                            (port + random.randint(-2 * n, 2 * n)))

    for port in ports:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = s.connect_ex((ip, port))
        s.close()
        if result != 0:
            return port
    raise ValueError("no open ports found")

def serve_and_open(html, ip='127.0.0.1', port=8888, n_retries=50, files=None,
                   ipython_warning=True):
    """Start a server serving the given HTML, and open a browser

    Parameters
    ----------
    html : string
        HTML to serve
    ip : string (default = '127.0.0.1')
        ip address at which the HTML will be served.
    port : int (default = 8888)
        the port at which to serve the HTML
    n_retries : int (default = 50)
        the number of nearby ports to search if the specified port is in use.
    files : dictionary (optional)
        dictionary of extra content to serve
    ipython_warning : bool (optional)
        if True (default), then print a warning if this is used within IPython
    """
    port = find_open_port(ip, port, n_retries)
    Handler = generate_handler(html, files)
    srvr = server.HTTPServer((ip, port), Handler)

    if ipython_warning:
        try:
            __IPYTHON__
        except:
            pass
        else:
            print(IPYTHON_WARNING)

    # Start the server
    print(("Serving to http://{0}:{1}/\n".format(ip, port) + 
           "[Ctrl-C to exit from terminal, or Ctrl-M i i to interrupt notebook kernel]"))
    sys.stdout.flush()

    # Use a thread to open a web browser pointing to the server
    b = lambda: webbrowser.open('http://{0}:{1}'.format(ip, port))
    threading.Thread(target=b).start()

    try:
        srvr.serve_forever()
    except (KeyboardInterrupt, SystemExit):
        print("\nstopping Server...")

    srvr.server_close()



# -*- coding: utf-8 -*-

from __future__ import print_function, unicode_literals

from escher.quick_server import serve_and_open
from escher.urls import get_url 
from escher.appdirs import user_cache_dir
from escher.generate_index import index

import os
from os.path import dirname, abspath, join, isfile, isdir
from warnings import warn
try:
    from urllib.request import urlopen
    from urllib.error import URLError
except:
    from urllib2 import urlopen, URLError
import json
import shutil
import re
from jinja2 import Environment, PackageLoader, Template
import codecs
import random
import string
from tornado.escape import url_escape

# set up jinja2 template location
env = Environment(loader=PackageLoader('escher', 'templates'))

# cache management

def get_cache_dir(name=None):
    """ Get the cache dir as a string.

    :param string name: An optional subdirectory within the cache

    """
    cache_dir = join(user_cache_dir('escher', appauthor='Zachary King'))
    if name is not None:
        cache_dir = join(cache_dir, name)
    try:
        os.makedirs(cache_dir)
    except OSError:
        pass
    return cache_dir

def clear_cache(different_cache_dir=None):
    """Empty the contents of the cache directory.

    :param string different_cache_dir: (Optional) The directory of another
                                       cache. This is mainly for testing.

    """
    if different_cache_dir is None:
        cache_dir = get_cache_dir()
    else:
        cache_dir = different_cache_dir    
        
    for root, dirs, files in os.walk(cache_dir):
        for f in files:
            os.unlink(join(root, f))
        for d in dirs:
            shutil.rmtree(join(root, d))

def local_index(cache_dir=get_cache_dir()):
    return index(cache_dir)

def list_cached_maps():
    """Return a list of all cached maps."""
    return local_index()['maps']

def list_cached_models():
    """Return a list of all cached models."""
    return local_index()['models']
    
# server management

def server_index():
    url = get_url('server_index', source='web', protocol='https')
    try:
        download = urlopen(url)
    except URLError:
        raise URLError('Could not contact Escher server')
    data = _decode_response(download)
    index = json.loads(data)
    return index

def list_available_maps():
    """Return a list of all maps available on the server"""
    return server_index()['maps']

def list_available_models():
    """Return a list of all models available on the server"""
    return server_index()['models']

# download maps and models

def _json_for_name(name, kind, cache_dir):
    # check the name
    name = name.replace('.json', '')

    def match_in_index(name, index, kind):
        return [(obj['organism'], obj[kind + '_name']) for obj in index[kind + 's']
                if obj[kind + '_name'] == name]
    
    # first check the local index
    match = match_in_index(name, local_index(cache_dir=cache_dir), kind)
    if len(match) == 0:
        path = None
    else:
        org, name = match[0]
        path = join(cache_dir, kind + 's', org, name + '.json')

    if path:
        # load the file
        with open(path, 'rb') as f:
            return f.read().decode('utf-8')
    # if the file is not present attempt to download
    else:
        try:
            index = server_index()
        except URLError:
            raise Exception(('Could not find the %s %s in the cache, and could '
                             'not connect to the Escher server' % (kind, name)))
        match = match_in_index(name, server_index(), kind)
        if len(match) == 0:
            raise Exception('Could not find the %s %s in the cache or on the server' % (kind, name))
        org, name = match[0]
        url = (get_url(kind + '_download', source='web', protocol='https') +
               '/'.join([url_escape(x, plus=False) for x in [org, name + '.json']]))
        warn('%s not in cache. Attempting download from %s' % (kind.title(), url))
        try:
            download = urlopen(url)
        except URLError:
            raise ValueError('No %s found in cache or at %s' % (kind, url))
        data = _decode_response(download)
        # save the file
        org_path = join(cache_dir, kind + 's', org)
        try:
            os.makedirs(org_path)
        except OSError:
            pass
        with open(join(org_path, name + '.json'), 'w') as outfile:
            outfile.write(data)
        return data

def model_json_for_name(model_name, cache_dir=get_cache_dir()):
    return _json_for_name(model_name, 'model', cache_dir)
   
def map_json_for_name(map_name, cache_dir=get_cache_dir()):
    return _json_for_name(map_name, 'map', cache_dir)

# helper functions

def _get_an_id():
    return (''.join(random.choice(string.ascii_lowercase)
                    for _ in range(10)))

def _decode_response(download):
    """Decode the urllib.response.addinfourl response."""
    data = download.read()
    try:                    # Python 2
        encoding = download.headers.getparam('charset')
    except AttributeError:  # Python 3
        encoding = download.headers.get_param('charset')
    if encoding:
        data = data.decode(encoding)
    else:
        data = data.decode('utf-8')
    return data

def _load_resource(resource, name, safe=False):
    """Load a resource that could be a file, URL, or json string."""
    # if it's a url, download it
    if resource.startswith('http://') or resource.startswith('https://'):
        try:
            download = urlopen(resource)
        except URLError as err:
            raise err
        else:
            return _decode_response(download)
    # if it's a filepath, load it
    if os.path.exists(resource):
        if (safe):
            raise Exception('Cannot load resource from file with safe mode enabled.')
        try:
            with open(resource, 'rb') as f:
                loaded_resource = f.read().decode('utf-8')
            _ = json.loads(loaded_resource)
        except ValueError as err:
            raise ValueError('%s not a valid json file' % name)
        else:
            return loaded_resource
    # try to validate the json
    try:
        _ = json.loads(resource)
    except ValueError as err:
        raise ValueError('Could not load %s. Not valid json, url, or filepath' % name)
    else:
        return resource
    raise Exception('Could not load %s.' % name)


class Builder(object):
    """A metabolic map that can be viewed, edited, and used to visualize data.
    
    This map will also show metabolic fluxes passed in during consruction.  It
    can be viewed as a standalone html inside a browswer. Alternately, the
    respresentation inside an IPython notebook will also display the map.

    Maps are stored in json files and are stored in a cache directory. Maps
    which are not found will be downloaded from a map repository if found.
    
    :param map_name:

        A string specifying a map to be downloaded from the Escher web server,
        or loaded from the cache.
    
    :param map_json:

        A JSON string, or a file path to a JSON file, or a URL specifying a JSON
        file to be downloaded.

    :param model: A Cobra model.

    :param model_name:

        A string specifying a model to be downloaded from the Escher web server,
        or loaded from the cache.
    
    :param model_json:

        A JSON string, or a file path to a JSON file, or a URL specifying a JSON
        file to be downloaded.

    :param embedded_css:

        The CSS (as a string) to be embedded with the Escher SVG.
    
    :param reaction_data:

        A dictionary with keys that correspond to reaction ids and values that
        will be mapped to reaction arrows and labels.

    :param metabolite_data:

        A dictionary with keys that correspond to metabolite ids and values that
        will be mapped to metabolite nodes and labels.

    :param gene_data:

        A dictionary with keys that correspond to gene ids and values that will
        be mapped to corresponding reactions.

    :param local_host:

        A hostname that will be used for any local files in dev mode.

    :param id:

        Specify an id to make the javascript data definitions unique. A random
        id is chosen by default.
    
    :param safe:

        If True, then loading files from the filesytem is not allowed. This is
        to ensure the safety of using Builder with a web server.

    **Keyword Arguments**

    These are defined in the Javascript API:

        - identifiers_on_map
        - show_gene_reaction_rules
        - unique_map_id
        - primary_metabolite_radius
        - secondary_metabolite_radius
        - marker_radius
        - hide_secondary_metabolites
        - reaction_styles
        - reaction_compare_style
        - reaction_scale
        - reaction_no_data_color
        - reaction_no_data_size
        - and_method_in_gene_reaction_rule
        - metabolite_styles
        - metabolite_compare_style
        - metabolite_scale
        - metabolite_no_data_color
        - metabolite_no_data_size
        - highlight_missing
        - allow_building_duplicate_reactions

    All keyword arguments can also be set on an existing Builder object
    using setter functions, e.g.:

    .. code:: python
    
        my_builder.set_reaction_styles(new_styles)

    """
    
    def __init__(self, map_name=None, map_json=None, model=None,
        model_name=None, model_json=None, embedded_css=None,
        reaction_data=None, metabolite_data=None, gene_data=None,
        local_host=None, id=None, safe=False, **kwargs):

        self.safe = safe
        
        # load the map
        self.map_name = map_name
        self.map_json = map_json
        self.loaded_map_json = None
        if map_name and map_json:
            warn('map_json overrides map_name')
        self._load_map()
        # load the model
        self.model = model
        self.model_name = model_name
        self.model_json = model_json
        self.loaded_model_json = None
        if sum([x is not None for x in (model, model_name, model_json)]) >= 2:
            warn('model overrides model_json, and both override model_name')
        self._load_model()
        # set the args
        self.reaction_data = reaction_data
        self.metabolite_data = metabolite_data
        self.gene_data = gene_data
        self.local_host = local_host
        
        # remove illegal characters from css
        try:
            self.embedded_css = (embedded_css.replace('\n', ''))
        except AttributeError:
            self.embedded_css = None
        # make the unique id
        self.the_id = _get_an_id() if id is None else id

        # set up the options
        self.options = ['identifiers_on_map',
                        'show_gene_reaction_rules',
                        'unique_map_id',
                        'primary_metabolite_radius',
                        'secondary_metabolite_radius',
                        'marker_radius',
                        'hide_secondary_metabolites',
                        'reaction_styles',
                        'reaction_compare_style',
                        'reaction_scale',
                        'reaction_no_data_color',
                        'reaction_no_data_size',
                        'and_method_in_gene_reaction_rule',
                        'metabolite_styles',
                        'metabolite_compare_style',
                        'metabolite_scale',
                        'metabolite_no_data_color',
                        'metabolite_no_data_size',
                        'highlight_missing',
                        'allow_building_duplicate_reactions']
        def get_getter_setter(o):
            """Use a closure."""
            # create local fget and fset functions 
            fget = lambda self: getattr(self, '_%s' % o)
            fset = lambda self, value: setattr(self, '_%s' % o, value)
            return fget, fset
        for option in self.options:
            fget, fset = get_getter_setter(option)
            # make the setter
            setattr(self.__class__, 'set_%s' % option, fset)
            # add property to self
            setattr(self.__class__, option, property(fget))
            # add corresponding local variable
            setattr(self, '_%s' % option, None)
            
        # set the kwargs
        for key, val in kwargs.items():
            try:
                getattr(self, 'set_%s' % key)(val)
            except AttributeError:
                print('Unrecognized keywork argument %s' % key)
        
    def _load_model(self):
        """Load the model.

        Try first from self.model, second from self.model_json, and
        third from from self.model_name.
        
        """
        if self.model is not None:
            try:
                import cobra.io
            except ImportError:
                raise Exception(('The COBRApy package must be available to load '
                                 'a COBRA model object'))
            self.loaded_model_json = cobra.io.to_json(self.model)
        elif self.model_json is not None:
            self.loaded_model_json = _load_resource(self.model_json,
                                                    'model_json',
                                                    safe=self.safe)
        elif self.model_name is not None:
            self.loaded_model_json = model_json_for_name(self.model_name)
                
    def _load_map(self):
        """Load the map from input map_json using _load_resource, or, secondarily,
           from map_name.

        """
        if self.map_json is not None:
            self.loaded_map_json = _load_resource(self.map_json,
                                                  'map_json',
                                                  safe=self.safe)
        elif self.map_name is not None:
            self.loaded_map_json = map_json_for_name(self.map_name)

    def _embedded_css(self, url_source):
        """Return a css string to be embedded in the SVG.

        Returns self.embedded_css if it has been assigned. Otherwise, attempts
        to download the css file.

        Arguments
        ---------
        
        url_source: Whether to load from 'web' or 'local'.
        

        """
        if self.embedded_css is not None:
            return self.embedded_css
            
        loc = get_url('builder_embed_css', source=url_source,
                      local_host=self.local_host, protocol='https')
        try:
            download = urlopen(loc)
        except URLError:
            raise Exception(('Could not find builder_embed_css. Be sure to pass '
                             'a local_host argument to Builder if js_source is dev or local '
                             'and you are in an iPython notebook or a static html file.'))
        data = _decode_response(download)
        return data.replace('\n', ' ')

    def _initialize_javascript(self, the_id, url_source):
        javascript = ("var map_data_{the_id} = {map_data};\n"
                      "var model_data_{the_id} = {model_data};\n"
                      "var reaction_data_{the_id} = {reaction_data};\n"
                      "var metabolite_data_{the_id} = {metabolite_data};\n"
                      "var gene_data_{the_id} = {gene_data};\n"
                      "var embedded_css_{the_id} = '{style}';\n").format(
                          the_id=the_id,
                          map_data=(self.loaded_map_json if self.loaded_map_json else
                                    'null'),
                          model_data=(self.loaded_model_json if self.loaded_model_json else
                                       'null'),
                          reaction_data=(json.dumps(self.reaction_data) if self.reaction_data else
                                         'null'),
                          metabolite_data=(json.dumps(self.metabolite_data) if self.metabolite_data else
                                           'null'),
                          gene_data=(json.dumps(self.gene_data) if self.gene_data else
                                           'null'),
                          style=self._embedded_css(url_source))
        return javascript

    def _draw_js(self, the_id, enable_editing, menu, enable_keys, dev,
                 fill_screen, scroll_behavior, never_ask_before_quit,
                 static_site_index_json): 
        draw = ("options = {{ enable_editing: {enable_editing},\n"
                "menu: {menu},\n"
                "enable_keys: {enable_keys},\n"
                "scroll_behavior: {scroll_behavior},\n"
                "fill_screen: {fill_screen},\n"
                "reaction_data: reaction_data_{the_id},\n"
                "metabolite_data: metabolite_data_{the_id},\n"
                "gene_data: gene_data_{the_id},\n"
                "never_ask_before_quit: {never_ask_before_quit},\n").format(
                    the_id=the_id,
                    enable_editing=json.dumps(enable_editing),
                    menu=json.dumps(menu),
                    enable_keys=json.dumps(enable_keys),
                    scroll_behavior=json.dumps(scroll_behavior),
                    fill_screen=json.dumps(fill_screen),
                    never_ask_before_quit=json.dumps(never_ask_before_quit))
        # Add the specified options
        for option in self.options:
            val = getattr(self, option)
            if val is None: continue
            draw = draw + "{option}: {value},\n".format(
                option=option,
                value=json.dumps(val)) 
        draw = draw + "};\n\n"
            
        # dev needs escher.
        dev_str = '' if dev else 'escher.'
        # parse the url in javascript, for building a static site
        if static_site_index_json:
            # get the relative location of the escher_download urls
            map_d = get_url('map_download', source='local')
            model_d = get_url('model_download', source='local')
            o = ('%sstatic.load_map_model_from_url("%s", "%s", \n%s, \noptions, function(map_data_%s, model_data_%s, options) {\n' %
                 (dev_str, map_d, model_d, static_site_index_json, the_id, the_id))
            draw = draw + o;
        # make the builder
        draw = draw + ('{dev_str}Builder(map_data_{the_id}, model_data_{the_id}, embedded_css_{the_id}, '
                       'd3.select("#{the_id}"), options);\n'.format(dev_str=dev_str, the_id=the_id))
        if static_site_index_json:
            draw = draw + '});\n'
        return draw
    
    def _get_html(self, js_source='web', menu='none', scroll_behavior='pan',
                  html_wrapper=False, enable_editing=False, enable_keys=False,
                  minified_js=True, fill_screen=False, height='800px',
                  never_ask_before_quit=False, static_site_index_json=None):
        """Generate the Escher HTML.

        Arguments
        --------

        js_source: Can be one of the following:
            'web' - (Default) use js files from escher.github.io.
            'local' - use compiled js files in the local escher installation. Works offline.
            'dev' - use the local, uncompiled development files. Works offline.

        menu: Menu bar options include:
            'none' - (Default) No menu or buttons.
            'zoom' - Just zoom buttons (does not require bootstrap).
            'all' - Menu and button bar (requires bootstrap).

        scroll_behavior: Scroll behavior options:
            'pan' - (Default) Pan the map.
            'zoom' - Zoom the map.
            'none' - No scroll events.

        minified_js: If True, use the minified version of js files. If
        js_source is 'dev', then this option is ignored.
            
        html_wrapper: If True, return a standalone html file.

        enable_editing: Enable the editing modes (build, rotate, etc.).

        enable_keys: Enable keyboard shortcuts.

        height: The height of the HTML container.

        never_ask_before_quit: Never display an alert asking if you want to
        leave the page. By default, this message is displayed if enable_editing
        is True.
        
        static_site_index_json: The index, as a JSON string, for the static
        site. Use javascript to parse the URL options. Used for
        generating static pages (see static_site.py).        

        """

        if js_source not in ['web', 'local', 'dev']:
            raise Exception('Bad value for js_source: %s' % js_source)
        
        if menu not in ['none', 'zoom', 'all']:
            raise Exception('Bad value for menu: %s' % menu)

        if scroll_behavior not in ['pan', 'zoom', 'none']:
            raise Exception('Bad value for scroll_behavior: %s' % scroll_behavior) 
            
        content = env.get_template('content.html')

        # if height is not a string
        if type(height) is int:
            height = "%dpx" % height
        elif type(height) is float:
            height = "%fpx" % height
        elif type(height) is str:
            height = str(height)
            
        # set the proper urls 
        url_source = 'local' if (js_source=='local' or js_source=='dev') else 'web'
        is_dev = (js_source=='dev')
        
        d3_url = get_url('d3', url_source, self.local_host)
        d3_rel_url = get_url('d3', url_source)
        escher_url = ('' if js_source=='dev' else
                      get_url('escher_min' if minified_js else 'escher',
                              url_source, self.local_host))
        jquery_url = ('' if not menu=='all' else
                      get_url('jquery', url_source, self.local_host))
        boot_css_url = ('' if not menu=='all' else
                        get_url('boot_css', url_source, self.local_host))
        boot_js_url = ('' if not menu=='all' else
                        get_url('boot_js', url_source, self.local_host))
        require_js_url = get_url('require_js', url_source, self.local_host)                     
        escher_css_url = get_url('builder_css', url_source, self.local_host)
        favicon_url = get_url('favicon', url_source, self.local_host)

        lh_string = ('' if self.local_host is None else
                     self.local_host.rstrip('/') + '/')
        
        html = content.render(id=self.the_id,
                              height=height,
                              dev=is_dev,
                              d3=d3_url,
                              d3_rel=d3_rel_url,
                              escher=escher_url,
                              jquery=jquery_url,
                              boot_css=boot_css_url,
                              boot_js=boot_js_url,
                              require_js=require_js_url,
                              escher_css=escher_css_url,
                              wrapper=html_wrapper,
                              title='Escher ' + ('Builder' if enable_editing else 'Viewer'),
                              favicon=favicon_url,
                              host=lh_string,
                              initialize_js=self._initialize_javascript(self.the_id, url_source),
                              draw_js=self._draw_js(self.the_id, enable_editing,
                                                    menu, enable_keys, is_dev,
                                                    fill_screen, scroll_behavior,
                                                    never_ask_before_quit,
                                                    static_site_index_json))
        return html

    def display_in_notebook(self, js_source='web', menu='zoom', scroll_behavior='none',
                            minified_js=True, height=500):
        """Embed the Map within the current IPython Notebook.

        :param string js_source:

            Can be one of the following:
            
            - *web* (Default) - Use JavaScript files from escher.github.io.
            - *local* - Use compiled JavaScript files in the local Escher installation. Works offline.
            - *dev* - Use the local, uncompiled development files. Works offline.
            
        :param string menu: Menu bar options include:
        
            - *none* - No menu or buttons.
            - *zoom* - Just zoom buttons.
            - Note: The *all* menu option does not work in an IPython notebook.

        :param string scroll_behavior: Scroll behavior options:
        
            - *pan* - Pan the map.
            - *zoom* - Zoom the map.
            - *none* - (Default) No scroll events.

        :param Boolean minified_js:

            If True, use the minified version of js files. If js_source is
            *dev*, then this option is ignored.

        :param height: Height of the HTML container.

        """
        html = self._get_html(js_source=js_source, menu=menu, scroll_behavior=scroll_behavior,
                              html_wrapper=False, enable_editing=False, enable_keys=False,
                              minified_js=minified_js, fill_screen=False, height=height,
                              never_ask_before_quit=True)
        if menu=='all':
            raise Exception("The 'all' menu option cannot be used in an IPython notebook.")
        # import here, in case users don't have requirements installed
        try:
            from IPython.display import HTML
        except ImportError:
            raise Exception('You need to be using the IPython notebook for this function to work')
        return HTML(html)

    
    def display_in_browser(self, ip='127.0.0.1', port=7655, n_retries=50, js_source='web',
                           menu='all', scroll_behavior='pan', enable_editing=True, enable_keys=True,
                           minified_js=True, never_ask_before_quit=False):
        """Launch a web browser to view the map.

        :param ip: The IP address to serve the map on.

        :param port:

            The port to serve the map on. If specified the port is occupied,
            then a random free port will be used.

        :param int n_retries:

            The number of times the server will try to find a port before
            quitting.
            
        :param string js_source:

            Can be one of the following:
            
            - *web* (Default) - Use JavaScript files from escher.github.io.
            - *local* - Use compiled JavaScript files in the local Escher installation. Works offline.
            - *dev* - Use the local, uncompiled development files. Works offline.

        :param string menu: Menu bar options include:
        
            - *none* - No menu or buttons.
            - *zoom* - Just zoom buttons.
            - *all* (Default) - Menu and button bar (requires Bootstrap).

        :param string scroll_behavior: Scroll behavior options:
        
            - *pan* - Pan the map.
            - *zoom* - Zoom the map.
            - *none* (Default) - No scroll events.
                
        :param Boolean enable_editing: Enable the map editing modes.

        :param Boolean enable_keys: Enable keyboard shortcuts.

        :param Boolean minified_js:

            If True, use the minified version of js files. If js_source is
            *dev*, then this option is ignored.

        :param Boolean never_ask_before_quit:

            Never display an alert asking if you want to leave the page. By
            default, this message is displayed if enable_editing is True.

        """
        html = self._get_html(js_source=js_source, menu=menu, scroll_behavior=scroll_behavior,
                              html_wrapper=True, enable_editing=enable_editing, enable_keys=enable_keys,
                              minified_js=minified_js, fill_screen=True, height="100%",
                              never_ask_before_quit=never_ask_before_quit)
        serve_and_open(html, ip=ip, port=port, n_retries=n_retries)
        
    def save_html(self, filepath=None, js_source='web', menu='all', scroll_behavior='pan',
                  enable_editing=True, enable_keys=True, minified_js=True,
                  never_ask_before_quit=False, static_site_index_json=None):
        """Save an HTML file containing the map.

        :param string filepath: The HTML file will be saved to this location.

        :param string js_source:

            Can be one of the following:
            
            - *web* (Default) - Use JavaScript files from escher.github.io.
            - *local* - Use compiled JavaScript files in the local Escher installation. Works offline.
            - *dev* - Use the local, uncompiled development files. Works offline.
            
        :param string menu: Menu bar options include:
        
            - *none* - No menu or buttons.
            - *zoom* - Just zoom buttons.
            - *all* (Default) - Menu and button bar (requires Bootstrap).

        :param string scroll_behavior: Scroll behavior options:
        
            - *pan* - Pan the map.
            - *zoom* - Zoom the map.
            - *none* (Default) - No scroll events.
            
        :param Boolean enable_editing: Enable the map editing modes.

        :param Boolean enable_keys: Enable keyboard shortcuts.

        :param Boolean minified_js:

            If True, use the minified version of js files. If js_source is
            *dev*, then this option is ignored.

        :param number height: Height of the HTML container.

        :param Boolean never_ask_before_quit:

            Never display an alert asking if you want to leave the page. By
            default, this message is displayed if enable_editing is True.

        :param string static_site_index_json:

            The index, as a JSON string, for the static site. Use javascript
            to parse the URL options. Used for generating static pages (see
            static_site.py).        

        """
        html = self._get_html(js_source=js_source, menu=menu, scroll_behavior=scroll_behavior,
                              html_wrapper=True, enable_editing=enable_editing, enable_keys=enable_keys,
                              minified_js=minified_js, fill_screen=True, height="100%",
                              never_ask_before_quit=never_ask_before_quit,
                              static_site_index_json=static_site_index_json)
        if filepath is not None:
            with open(filepath, 'wb') as f:
                f.write(html.encode('utf-8'))
            return filepath
        else:
            from tempfile import mkstemp
            from os import write, close
            os_file, filename = mkstemp(suffix=".html", text=False) # binary
            write(os_file, html.encode('utf-8'))
            close(os_file)
            return filename
